@@ -1,37 +1,44 @@
-;;; geiser-gauche.el -- Gauche Scheme's implementation of the geiser protocols
+;;; geiser-gauche.el --- Gauche scheme support for Geiser -*- lexical-binding:t -*-
 
-;; This program is free software; you can redistribute it and/or
-;; modify it under the terms of the Modified BSD License. You should
-;; have received a copy of the license along with this program. If
-;; not, see <http://www.xfree86.org/3.3.6/COPYRIGHT2.html#5>.
+;; Copyright (C) 2020 András Simonyi
+
+;; Author: András Simonyi <andras.simonyi@gmail.com>
+;; Maintainer: András Simonyi <andras.simonyi@gmail.com>
+;; URL: https://gitlab.com/emacs-geiser/gauche
+;; Keywords: languages, gauche, scheme, geiser
+;; Package-Requires: ((emacs "26.1") (geiser "0.11.2"))
+;; Package-Version: 20251213.1148
+;; Package-Revision: adbd3c8d9031
+;; SPDX-License-Identifier: BSD-3-Clause
+
+;; This file is not part of GNU Emacs.
+
+;;; Commentary:
+
+;; geiser-gauche adds Gauche Scheme support to the `geiser' package
 
-
 ;;; Code:
 
-(require 'geiser-connection)
+(require 'info-look)
+(require 'cl-lib)
+(require 'compile)
+
 (require 'geiser-syntax)
 (require 'geiser-custom)
-(require 'geiser-base)
 (require 'geiser-eval)
-(require 'geiser-edit)
 (require 'geiser-log)
+(require 'geiser-impl)
 (require 'geiser)
 
-(require 'compile)
-(require 'info-look)
-
-(eval-when-compile (require 'cl-lib))
-
 
-;;; Customization:
+;;; Customization
 
 (defgroup geiser-gauche nil
-  "Customization for Geiser's Gauche Scheme flavour."
+  "Customization for Geiser's Gauche flavour."
   :group 'geiser)
 
-(geiser-custom--defcustom geiser-gauche-binary
-    "gosh"
-  "Name to use to call the Gauche Scheme executable when starting a REPL."
+(geiser-custom--defcustom geiser-gauche-binary "gosh"
+  "Name to use to call the Gauche executable when starting a REPL."
   :type '(choice string (repeat string))
   :group 'geiser-gauche)
 
@@ -60,98 +67,36 @@
   "Non-nil means pop up the Info buffer in another window."
   :type 'boolean
   :group 'geiser-gauche)
-
-;;; REPL support:
 
-(defun geiser-gauche--binary ()
-  (if (listp geiser-gauche-binary)
-      (car geiser-gauche-binary)
-    geiser-gauche-binary))
-
-(defun geiser-gauche--parameters ()
-  "Return a list with all parameters needed to start Gauche Scheme."
-  `(,@geiser-gauche-extra-command-line-parameters
-    "-l" ,(expand-file-name "gauche/geiser.scm" geiser-scheme-dir)
-    ,@(and (listp geiser-gauche-binary) (cdr geiser-gauche-binary))))
-
-(defconst geiser-gauche--prompt-regexp "gosh\\(\\[[^(]+\\]\\)?> ") 
+(geiser-custom--defcustom geiser-gauche-rnrs nil
+  "Non-nil specifies rnrs version, corresponding to gosh's -r option"
+  :type 'integer
+  :group 'geiser-gauche)
 
 
-;;; Evaluation support:
+;;; Utils
 
-(defun geiser-gauche--geiser-procedure (proc &rest args)
-  ;; (with-current-buffer "*scratch*"
-  ;;   (goto-char (point-max))
-  ;;   (insert (format "\nGeiser PROC: %s, ARGS: %s \n" proc args)))
-  (cl-case proc
-    ;; Autodoc and symbol-location makes use of the {{cur-module}} cookie to
-    ;; pass current module information
-    ((autodoc symbol-location)
-     (format "(eval '(geiser:%s %s {{cur-module}}) (find-module 'geiser))"
-	     proc (mapconcat 'identity args " ")))
-    ;; Eval and compile are (module) context sensitive
-    ((eval compile)
-     (let ((module (cond ((string-equal "'()" (car args))
-			  "'()")
-			 ((and (car args))
-			  (concat "'" (car args)))
-			 (t
-			  "#f")))
-	   (form (mapconcat 'identity (cdr args) " ")))
-       ;; The {{cur-module}} cookie is replaced by the current module for
-       ;; commands that need it
-       (replace-regexp-in-string
-	"{{cur-module}}"
-	(if (string= module "'#f")
-	    (format "'%s" (geiser-gauche--get-module))
-	  module)
-	(format "(eval '(geiser:eval %s '%s) (find-module 'geiser))" module form))))
-    ;; The rest of the commands are all evaluated in the geiser module 
-    (t
-     (let ((form (mapconcat 'identity args " ")))
-       (format "(eval '(geiser:%s %s) (find-module 'geiser))" proc form)))))
+(defconst geiser-gauche--load-dir (file-name-directory load-file-name)
+  "The directory from which geiser-gauche was loaded.")
 
-(defconst geiser-gauche--module-re
-  "(define-module +\\([[:alnum:].-]+\\)")
+(defun geiser-gauche--symbol-begin (_module)
+  "Return the beginning position of the symbol at point."
+  (save-excursion (skip-syntax-backward "^'-()>") (point)))
 
-(defun geiser-gauche--get-current-repl-module ()
-  (substring 
-   (geiser-eval--send/result
-    '(list (list (quote result) (write-to-string (current-module)))))
-   9 -1))
+
+;;; Guess whether buffer is Gauche
 
-(defun geiser-gauche--get-module (&optional module)
-  (cond ((null module)
-         (save-excursion
-           (geiser-syntax--pop-to-top)
-           (if (or (re-search-backward geiser-gauche--module-re nil t)
-                   (looking-at geiser-gauche--module-re)
-                   (re-search-forward geiser-gauche--module-re nil t))
-               (geiser-gauche--get-module (match-string-no-properties 1))
-	     ;; Return the repl module as fallback
-             (geiser-gauche--get-module
-	      (geiser-gauche--get-current-repl-module)))))
-	((symbolp module) module)
-        ((listp module) module)
-        ((stringp module)
-         (condition-case nil
-             (car (geiser-syntax--read-from-string module))
-           (error :f)))
-        (t :f)))
+(defconst geiser-gauche--guess-re
+  (regexp-opt '("gauche" "gosh")))
 
-(defun geiser-gauche--enter-command (module)
-  (format "(select-module %s)" module))
+(defun geiser-gauche--guess ()
+  "Guess whether the current buffer edits Gauche code or REPL."
+  (save-excursion
+    (goto-char (point-min))
+    (re-search-forward geiser-gauche--guess-re nil t)))
 
-(defun geiser-gauche--import-command (module)
-  (format "(import %s)" module))
-
-(defun geiser-gauche--symbol-begin (module)
-  (if module
-      (max (save-excursion (beginning-of-line) (point))
-           (save-excursion (skip-syntax-backward "^(>") (1- (point))))
-    (save-excursion (skip-syntax-backward "^'-()>") (point))))
-
-(defun geiser-gauche--exit-command () "(exit)")
+
+;;; Keywords and syntax
 
 (defconst geiser-gauche--binding-forms
   '("and-let" "and-let1" "let1" "if-let1" "rlet1" "receive" "fluid-let" "let-values"
@@ -161,7 +106,7 @@
 (defconst geiser-gauche--binding-forms*
   '("and-let*" "let*-values" ))
 
-(defconst geiser-gauche-builtin-keywords
+(defconst geiser-gauche--builtin-keywords
   '("and-let"
     "and-let1"
     "assume"
@@ -217,31 +162,152 @@
     "with-time-counter"))
 
 (defun geiser-gauche--keywords ()
+  "Add geiser-specific keywords to the default ones."
   (append
    (geiser-syntax--simple-keywords geiser-gauche-extra-keywords)
-   (geiser-syntax--simple-keywords geiser-gauche-builtin-keywords)))
+   (geiser-syntax--simple-keywords geiser-gauche--builtin-keywords)))
 
 (geiser-syntax--scheme-indent
  (case-lambda: 0)
  (let1 1))
 
 
-;;; REPL startup
+;;; REPL
 
-(defconst geiser-gauche-minimum-version "0.9.9")
+(defconst geiser-gauche--prompt-regexp "gosh\\(\\[[^(]+\\]\\)?> ")
+
+(defconst geiser-gauche--minimum-version "0.9.6")
+
+(defconst geiser-gauche--exit-command
+  "(exit)")
+
+(defun geiser-gauche--binary ()
+  "Return the runnable Gauche binary name without path."
+  (if (listp geiser-gauche-binary)
+      (car geiser-gauche-binary)
+    geiser-gauche-binary))
+
+(defun geiser-gauche--parameters ()
+  "Return a list with all parameters needed to start Gauche Scheme."
+  `(,@geiser-gauche-extra-command-line-parameters
+    ,@(if geiser-gauche-rnrs
+           (list "-r" (number-to-string geiser-gauche-rnrs)
+                 "-l" (expand-file-name "geiser-gauche-r7rs.scm"
+                                        geiser-gauche--load-dir))
+        (list "-l" (expand-file-name "geiser-gauche.scm"
+                                     geiser-gauche--load-dir)))
+    ,@(and (listp geiser-gauche-binary) (cdr geiser-gauche-binary))))
 
 (defun geiser-gauche--version (binary)
-  (cadr (read (cadr (process-lines "gosh" "-V")))))
+  "Return the version of a Gauche BINARY."
+  (let ((v (process-lines binary "-V")))
+    (if (< 1 (length v))
+	;; Multiline version info is (hopefully) a sexp.
+	(cadr (read (cadr v)))
+      ;; One line is a free-text version description string.
+      (let ((s (car v))
+	    (start (string-match "version" (car v))))
+	(substring s (+ start 8) (+ start 13))))))
 
-(defun geiser-gauche--startup (remote)
+(defun geiser-gauche--startup (_remote)
+  "Initialize a Gauche REPL."
   (let ((geiser-log-verbose-p t))
     (compilation-setup t)
     (geiser-eval--send/wait "(newline)")))
 
+
+;;; Evaluation
+
+(defun geiser-gauche--geiser-procedure (proc &rest args)
+  "String to send to the REPL to execute geiser PROC with ARGS."
+  (cl-case proc
+    ;; Autodoc and symbol-location makes use of the {{cur-module}} cookie to
+    ;; pass current module information
+    ((autodoc symbol-location completions)
+     (format "(eval '(geiser:%s %s {{cur-module}}) (find-module 'geiser))"
+	     proc (mapconcat #'identity args " ")))
+    ;; Eval and compile are (module) context sensitive
+    ((eval compile)
+     (let ((module (cond ((string-equal "'()" (car args))
+			  "'()")
+			 ((and (car args))
+			  (concat "'" (car args)))
+			 (t
+			  "#f")))
+	   (form (mapconcat #'identity (cdr args) " ")))
+       ;; The {{cur-module}} cookie is replaced by the current module for
+       ;; commands that need it
+       (replace-regexp-in-string
+	"{{cur-module}}"
+	(if (string= module "'#f")
+	    (let ((cur-module (geiser-gauche--get-module)))
+	      (if (eq cur-module :f)
+		  "'#f"
+		(format "'%s" cur-module)))
+	  module)
+	(format "(eval '(geiser:eval %s '%s) (find-module 'geiser))" module form))))
+    ;; The rest of the commands are all evaluated in the geiser module
+    (t
+     (let ((form (mapconcat #'identity args " ")))
+       (format "(eval '(geiser:%s %s) (find-module 'geiser))" proc form)))))
+
+(defun geiser-gauche--eval-bounds ()
+  "Return the bounds for `geiser-eval-buffer' command."
+  (cons (if (equal "#!" (buffer-substring (point-min) (+ 2 (point-min))))
+            (save-excursion
+              (goto-char (point-min))
+              (forward-line 1)
+              (point))
+          (point-min))
+        (point-max)))
+
+
+;;; Module handling
+
+(defconst geiser-gauche--module-re
+  "(define-module +\\([[:alnum:].-]+\\)"
+  "Regex for locating the module defined in a scheme source.")
+
+(defun geiser-gauche--get-current-repl-module ()
+  "Return the current REPL module's name as a string."
+  (substring
+   (geiser-eval--send/result
+    '(list (list (quote result) (write-to-string (current-module)))))
+   9 -1))
+
+(defun geiser-gauche--get-module (&optional module)
+  "Return the current module as a symbol.
+If optional MODULE is non-nil then return its normalized symbol
+form."
+  (cond ((null module)
+         (save-excursion
+           (geiser-syntax--pop-to-top)
+           (if (or (re-search-backward geiser-gauche--module-re nil t)
+                   (looking-at geiser-gauche--module-re)
+                   (re-search-forward geiser-gauche--module-re nil t))
+               (geiser-gauche--get-module (match-string-no-properties 1))
+             :f)))
+	((symbolp module) module)
+        ((listp module) module)
+        ((stringp module)
+         (condition-case nil
+             (car (geiser-syntax--read-from-string module))
+           (error :f)))
+        (t :f)))
+
+(defun geiser-gauche--enter-command (module)
+  "Return a scheme expression string to enter MODULE."
+  (format "(select-module %s)" module))
+
+(defun geiser-gauche--import-command (module)
+  "Return a scheme expression string to import MODULE."
+  (format "(import %s)" module))
+
 
 ;;; Error display
 
-(defun geiser-gauche--display-error (module key msg)
+(defun geiser-gauche--display-error (_module key msg)
+  "Display evaluation error information KEY and MSG."
   (when key
     (insert key)
     (save-excursion
@@ -253,10 +319,11 @@
   (if (and msg (string-match "\\(.+\\)$" msg)) (match-string 1 msg) key))
 
 
-;;; Manual look up
-;;; code adapted from the Guile implementation
+;;; Manual lookup
+;; code adapted from the Guile implementation
 
 (defun geiser-gauche--info-spec (&optional nodes)
+  "Return an info docspec list for NODES."
   (let* ((nrx "^[       ]+-+ [^:]+:[    ]*")
          (drx "\\b")
          (res (when (Info-find-file "gauche-refe" t)
@@ -273,7 +340,8 @@
                       :regexp "[^()`',\"        \n]+"
                       :doc-spec (geiser-gauche--info-spec))
 
-(defun geiser-gauche--manual-look-up (id mod)
+(defun geiser-gauche--manual-look-up (id _mod)
+  "Look up ID in the Gauche info manual."
   (let ((info-lookup-other-window-flag
          geiser-gauche-manual-lookup-other-window-p))
     (info-lookup-symbol (symbol-name id) 'geiser-gauche-mode))
@@ -282,30 +350,17 @@
   (search-forward (format "%s" id) nil t))
 
 
-;;; Guess whether buffer is Gauche
-
-(defconst geiser-gauche--guess-re
-  (regexp-opt '("gauche" "gosh")))
-
-(defun geiser-gauche--guess ()
-  (save-excursion
-    (goto-char (point-min))
-    (re-search-forward geiser-gauche--guess-re nil t)))
-
-
-;;; Implementation definition:
+;;; Implementation definition
 
 (define-geiser-implementation gauche
   (unsupported-procedures '(callers callees generic-methods))
   (binary geiser-gauche--binary)
   (arglist geiser-gauche--parameters)
   (version-command geiser-gauche--version)
-  (minimum-version geiser-gauche-minimum-version)
+  (minimum-version geiser-gauche--minimum-version)
   (repl-startup geiser-gauche--startup)
   (prompt-regexp geiser-gauche--prompt-regexp)
   (debugger-prompt-regexp nil)
-  ;; geiser-gauche--debugger-prompt-regexp
-  ;; (enter-debugger geiser-gauche--enter-debugger)
   (marshall-procedure geiser-gauche--geiser-procedure)
   (find-module geiser-gauche--get-module)
   (enter-command geiser-gauche--enter-command)
@@ -318,10 +373,24 @@
   (external-help geiser-gauche--manual-look-up)
   (check-buffer geiser-gauche--guess)
   (keywords geiser-gauche--keywords)
-  (case-sensitive geiser-gauche-case-sensitive-p))
+  (case-sensitive geiser-gauche-case-sensitive-p)
+  (eval-bounds geiser-gauche--eval-bounds))
 
-(geiser-impl--add-to-alist 'regexp "\\.scm$" 'gauche t)
+(geiser-implementation-extension 'gauche "scm")
 
-(add-to-list 'geiser-active-implementations 'gauche)
 
+;;; Autoloads
+
+;;;###autoload
+(geiser-activate-implementation 'gauche)
+
+;;;###autoload
+(autoload 'run-gauche "geiser-gauche" "Start a Geiser Gauche Scheme REPL." t)
+
+;;;###autoload
+(autoload 'switch-to-gauche "geiser-gauche"
+  "Start a Geiser Gauche Scheme REPL, or switch to a running one." t)
+
 (provide 'geiser-gauche)
+
+;;; geiser-gauche.el ends here
